# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dot_general."""
import copy
from typing import Any, Dict, Iterable, Optional, Sequence

from absl.testing import absltest
from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.jax import aqt_ops
from aqt.jax import aqt_tensor
from aqt.test import aqt_test_shared_base
from flax import linen as nn
import jax
from jax import lax
import jax.numpy as jnp
import numpy as np


# pylint: disable=g-long-lambda


def test_stats_config(share_stats_axes):
  """Config base for all the experiments."""
  return aqt_config.StatsConfig(
      ema_update_count=1,  # single update overwrites the stats
      update_count_prior=0,  # easier equations
      share_stats_axes=share_stats_axes,  # one stat per whole tensor
      tpu_cross_replica_sum=False,  # no cross-tpu reduction
      filter_zeros=False,  # on default zeros are a number like any other
  )


def calibration_config(const_coeff: float) -> aqt_config.CalibrationConfig:
  return aqt_config.CalibrationConfig(const_bound_coeff=const_coeff)


def config_from_schedule(schedule, share_stats_axes):
  """Generates a schedule config from [(start, end, bits), ...]."""
  tensor_configs = []
  for start, end, bits in schedule:
    int_quant_config = aqt_config.IntQuantConfig(bits, preserve_zero=False)
    tensor_config = aqt_config.AqtTensorConfig(
        freeze_scale_at_begin=True,
        quant_config=int_quant_config,
        begin_at_event=start,
        end_at_event=end,
        calibration_config=calibration_config(1.0))
    tensor_configs.append(tensor_config)
  return aqt_config.AqtScheduleConfig(
      test_stats_config(share_stats_axes), tensor_configs)


class DotModule(nn.Module):
  lhs_config: Optional[aqt_config.AqtScheduleConfig]
  rhs_config: Optional[aqt_config.AqtScheduleConfig]
  lhs_shape: Iterable[int]
  rhs_shape: Iterable[int]

  @nn.compact
  def __call__(self):
    lhs_quantizer = aqt_tensor.TensorQuantizer(list(self.lhs_shape),
                                               self.lhs_config)
    rhs_quantizer = aqt_tensor.TensorQuantizer(list(self.rhs_shape),
                                               self.rhs_config)
    return lambda lhs, rhs: aqt_ops.aqt_dot(
        lhs,
        rhs,
        lhs_quantizer,
        rhs_quantizer)


class DotGeneralModule(nn.Module):
  lhs_config: Optional[aqt_config.AqtScheduleConfig]
  rhs_config: Optional[aqt_config.AqtScheduleConfig]
  lhs_shape: Iterable[int]
  rhs_shape: Iterable[int]

  @nn.compact
  def __call__(self, lhs, rhs, train=True):
    lhs_quantizer = aqt_tensor.TensorQuantizer(list(self.lhs_shape),
                                               self.lhs_config)
    rhs_quantizer = aqt_tensor.TensorQuantizer(list(self.rhs_shape),
                                               self.rhs_config)

    lhs_quantizer.update(lhs, None, 0)  # pytype: disable=wrong-arg-types  # jax-ndarray
    rhs_quantizer.update(rhs, None, 0)  # pytype: disable=wrong-arg-types  # jax-ndarray

    return lambda lhs, rhs, dimension_numbers: aqt_ops.aqt_dot_general(
        lhs,
        rhs,
        lhs_quantizer,
        rhs_quantizer,
        dimension_numbers=dimension_numbers,
        train=train)


def _generate_dimension_numbers() -> Sequence[Dict[str, Any]]:
  """Generates arbitrary dimension numbers for a tensor of shape (2, 2, 2)."""
  keys = ["testcase_name", "dimension_numbers"]
  # ((lhs_contracting_dims, rhs_contracting_dims), (lhs_batch_dims,
  # rhs_batch_dims))
  cases = [
      ("batch_matmul", (((2,), (1,)), ((0,), (0,)))),
      ("one_cont_two_batch_dims", (((2,), (2,)), ((0, 1,), (0, 1,)))),
      ("two_cont_one_batch_dims", (((1, 2), (1, 2)), ((0,), (0,)))),
      ("one_contracting_dims", (((2,), (1,)), ((), ()))),
      ("two_contracting_dims", (((1, 2), (1, 2)), ((), ()))),
  ]
  return [dict(zip(keys, vals)) for vals in cases]


class AqtDotGeneralTest(parameterized.TestCase):

  def vgrad(self, f, lhs, rhs):
    y, vjp_fn = jax.vjp(f, lhs, rhs)
    return vjp_fn(jnp.ones(y.shape))

  def get_dot_module(self, lhs_config, rhs_config, lhs_shape, rhs_shape):
    module = DotModule(lhs_config, rhs_config, lhs_shape, rhs_shape)
    return module.init_with_output(jax.random.PRNGKey(0))

  def get_dot_general_module(self, lhs_config, rhs_config, lhs, rhs):
    module = DotGeneralModule(lhs_config, rhs_config, lhs.shape, rhs.shape)
    return module.init_with_output(jax.random.PRNGKey(0), lhs, rhs)

  def exact_int8_dot_general_example(self,
                                     lhs_use_quantized_variable=False,
                                     rhs_use_quantized_variable=False):
    lhs_config, lhs, rhs_config, rhs = aqt_test_shared_base.exact_int8_example(
        lhs_shape=(2, 2, 2),
        rhs_shape=(2, 2, 2),
        lhs_share_stats_axes=[0, 1, 2],
        rhs_share_stats_axes=[0, 1, 2],
        lhs_use_quantized_variable=lhs_use_quantized_variable,
        rhs_use_quantized_variable=rhs_use_quantized_variable)
    return lhs_config, jnp.array(lhs), rhs_config, jnp.array(rhs)

  def test_dot_none(self):
    no_quant_config = aqt_config.AqtScheduleConfig(
        test_stats_config([0, 1]), [])
    lhs = np.random.uniform(-1.0, 1.0, size=(2, 3)).astype(np.float32)
    rhs = np.random.uniform(-1.0, 1.0, size=(3, 4)).astype(np.float32)
    dot, _ = self.get_dot_module(no_quant_config, no_quant_config, lhs.shape,
                                 rhs.shape)
    actual_ret = dot(lhs, rhs)
    expected_ret = lax.dot(lhs, rhs)
    np.testing.assert_array_equal(actual_ret, expected_ret)

  def test_dot_incompatible_shapes(self):
    config = aqt_config.AqtScheduleConfig(
        test_stats_config([0, 1]), [])
    lhs = np.random.uniform(-1.0, 1.0, size=(2, 3)).astype(np.float32)
    rhs = np.random.uniform(-1.0, 1.0, size=(2, 4)).astype(np.float32)
    dot, _ = self.get_dot_module(config, config, lhs.shape, rhs.shape)
    # the last dimension of lhs and the first dimension of rhs must be the same.
    with self.assertRaisesRegex(TypeError, "Incompatible shapes for dot"):
      dot(lhs, rhs)

    lhs_rank3 = np.random.uniform(-1.0, 1.0, size=(2, 3, 2)).astype(np.float32)
    dot, _ = self.get_dot_module(config, config, lhs_rank3.shape, rhs.shape)
    # lhs and rhs must be rank1 or rank2.
    with self.assertRaisesRegex(TypeError, "Incompatible shapes for dot"):
      dot(lhs_rank3, rhs)

    rhs_rank3 = np.random.uniform(-1.0, 1.0, size=(3, 4, 5)).astype(np.float32)
    dot, _ = self.get_dot_module(config, config, lhs.shape, rhs_rank3.shape)
    # lhs and rhs must be rank1 or rank2.
    with self.assertRaisesRegex(TypeError, "Incompatible shapes for dot"):
      dot(lhs, rhs_rank3)

  @parameterized.named_parameters(_generate_dimension_numbers())
  def test_dot_general_none(self, dimension_numbers):
    no_quant_config = aqt_config.AqtScheduleConfig(
        test_stats_config([0, 1, 2]), [])
    lhs = np.random.uniform(-1.0, 1.0, size=(2, 2, 2)).astype(np.float32)
    rhs = np.random.uniform(-1.0, 1.0, size=(2, 2, 2)).astype(np.float32)

    # Creates a dot_general module with empty tensor configs.
    dot_general, _ = self.get_dot_general_module(no_quant_config,
                                                 no_quant_config, lhs, rhs)
    actual_ret = dot_general(lhs, rhs, dimension_numbers=dimension_numbers)
    expected_ret = lax.dot_general(
        lhs, rhs, dimension_numbers=dimension_numbers)
    np.testing.assert_array_equal(actual_ret, expected_ret)

    # Creates a dot_general module with None schedule configs.
    dot_general, _ = self.get_dot_general_module(None, None, lhs, rhs)
    actual_ret = dot_general(lhs, rhs, dimension_numbers=dimension_numbers)
    np.testing.assert_array_equal(actual_ret, expected_ret)

  @parameterized.named_parameters(_generate_dimension_numbers())
  def test_validates_contraction(self, dimension_numbers):
    lhs_config, lhs, rhs_config, rhs = self.exact_int8_dot_general_example()
    bad_rhs_config = copy.deepcopy(rhs_config)
    bad_rhs_config.stats_config.share_stats_axes = [0]
    dot_general, _ = self.get_dot_general_module(lhs_config, bad_rhs_config,
                                                 lhs, rhs)
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "expected rhs dot_general contraction axis"):
      dot_general(lhs, rhs, dimension_numbers=dimension_numbers)
    bad_lhs_config = copy.deepcopy(lhs_config)
    bad_lhs_config.stats_config.share_stats_axes = [0]
    dot_general, _ = self.get_dot_general_module(bad_lhs_config, rhs_config,
                                                 lhs, rhs)
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "expected lhs dot_general contraction axis"):
      dot_general(lhs, rhs, dimension_numbers=dimension_numbers)

  @parameterized.named_parameters(
      aqt_test_shared_base.generate_unaligned_schedule_intervals())
  def test_unaligned_schedule_intervals(self, lhs_intervals, rhs_intervals):
    bits = 8
    lhs_intervals = [(start, stop, bits) for start, stop in lhs_intervals]
    rhs_intervals = [(start, stop, bits) for start, stop in rhs_intervals]
    lhs_config = config_from_schedule(lhs_intervals, [0, 1])
    rhs_config = config_from_schedule(rhs_intervals, [0, 1])
    lhs = rhs = np.ones((1, 1)).astype(np.float32)
    lhs = jnp.array(lhs)
    rhs = jnp.array(lhs)
    dot_general, _ = self.get_dot_general_module(lhs_config, rhs_config, lhs,
                                                 rhs)
    dimension_numbers = (((1,), (0,)), ((), ()))
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "intervals do not match|config len"):
      dot_general(lhs, rhs, dimension_numbers=dimension_numbers)

  def get_dot_general_for_grad(self, lhs_config, rhs_config, lhs, rhs,
                               dimension_numbers):
    # jax.vjp does not take a tuple of primals while `dimension_numbers` is in a
    # form of tuple. Thus, this method returns a lambda with `dimension_numbers`
    # argument which will be called for jax.vjp to get gradients.
    dot_general, _ = self.get_dot_general_module(lhs_config, rhs_config, lhs,
                                                 rhs)
    return lambda lhs, rhs: dot_general(lhs, rhs, dimension_numbers)

  @parameterized.named_parameters(_generate_dimension_numbers())
  def test_vars_dont_kill_grads(self, dimension_numbers):
    lhs_config_novar, lhs, rhs_config_novar, rhs = self.exact_int8_dot_general_example(
    )
    lhs_config, _, rhs_config, _ = self.exact_int8_dot_general_example(
        lhs_use_quantized_variable=True, rhs_use_quantized_variable=True)

    unsaved_dot_general = self.get_dot_general_for_grad(lhs_config_novar,
                                                        rhs_config_novar, lhs,
                                                        rhs, dimension_numbers)
    saved_dot_general = self.get_dot_general_for_grad(lhs_config, rhs_config,
                                                      lhs, rhs,
                                                      dimension_numbers)

    saved_grads = self.vgrad(saved_dot_general, lhs, rhs)
    unsaved_grads = self.vgrad(unsaved_dot_general, lhs, rhs)
    zipped_grads = zip(saved_grads, unsaved_grads)
    for actual_grad, expected_grad in zipped_grads:
      np.testing.assert_array_equal(actual_grad, expected_grad)

  @parameterized.named_parameters(_generate_dimension_numbers())
  def test_vars_over_inputs_at_inference(self, dimension_numbers):
    lhs_config, lhs, rhs_config, rhs = self.exact_int8_dot_general_example(
        lhs_use_quantized_variable=True, rhs_use_quantized_variable=True)
    module = DotGeneralModule(lhs_config, rhs_config, lhs.shape, rhs.shape)

    # Since quantized variables are not used at training, the dot_general with
    # zero inputs should produce zero values.
    dot_general, state = module.init_with_output(
        jax.random.PRNGKey(0), lhs, rhs)
    actual = dot_general(
        jnp.zeros_like(lhs), jnp.zeros_like(rhs), dimension_numbers)
    expected = jnp.zeros_like(actual)
    np.testing.assert_array_equal(actual, expected)

    # Since quantized variables should be always used at inference, the
    # dot_general will rely on quantized variables.
    dot_general_infer, _ = module.apply(state, lhs, rhs, False, mutable=True)
    actual = dot_general_infer(
        jnp.zeros_like(lhs), jnp.zeros_like(rhs), dimension_numbers)
    expected = dot_general(lhs, rhs, dimension_numbers)
    np.testing.assert_array_equal(actual, expected)

  @parameterized.named_parameters(_generate_dimension_numbers())
  def test_exact_grads(self, dimension_numbers):
    lhs_config, lhs, rhs_config, rhs = self.exact_int8_dot_general_example()

    aqt_dot_general = self.get_dot_general_for_grad(lhs_config, rhs_config, lhs,
                                                    rhs, dimension_numbers)

    aqt_dg = lambda lhs, rhs: jnp.sum(aqt_dot_general(lhs, rhs)**2)
    aqt_dg_grads = self.vgrad(aqt_dg, lhs, rhs)

    dg = lambda lhs, rhs: jnp.sum(lax.dot_general(
        lhs, rhs, dimension_numbers=dimension_numbers)**2)
    dg_grads = self.vgrad(dg, lhs, rhs)
    for aqt_grad, jax_grad in zip(aqt_dg_grads, dg_grads):
      actual = aqt_grad
      expected = jax_grad
      np.testing.assert_array_equal(actual, expected)

  def test_weight_only_quantization(self):
    lhs = np.random.uniform(-1.0, 1.0, size=(2, 2, 2)).astype(np.float32)
    rhs = np.random.uniform(-1.0, 1.0, size=(2, 2, 2)).astype(np.float32)
    lhs_config = config_from_schedule([(None, None, 8)], [0, 1, 2])
    rhs_config = config_from_schedule([(None, None, 8)], [0, 1, 2])
    lhs_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()
    # dimension numbers for batch matmul
    dimension_numbers = (((2,), (1,)), ((0,), (0,)))

    dot_general, _ = self.get_dot_general_module(lhs_config, rhs_config, lhs,
                                                 rhs)
    result_with_float_config = dot_general(
        lhs, rhs, dimension_numbers=dimension_numbers)
    dot_general, _ = self.get_dot_general_module(None, rhs_config, lhs, rhs)
    result_with_none_config = dot_general(
        lhs, rhs, dimension_numbers=dimension_numbers)
    np.testing.assert_array_equal(result_with_float_config,
                                  result_with_none_config)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Primitives for neural network quantization implemented in jax."""
from typing import Any, Iterable, Optional, TypeVar

import jax
import jax.numpy as jnp

# Custom type for bounds for type hinting
BoundsT = TypeVar('BoundsT', float, jnp.ndarray)

jnp_dtype = Any  # type of jax dtype; b/164524367 use more specific type.

# Global bool to control the use of epsilon in the denominator of the scaling
# methods signed_int_scale and unsigned_int_scale. Epsilon is added to avoid
# division by 0. For testing, one may choose to disable the epsilon by setting
# this global to True.
# As this is a global variable, please modify it only before calling any
# functions that use it.
DISABLE_EPSILON_IN_SCALE_FUN_FOR_TESTING = False


def add_straight_through_estimator(jax_function) -> None:
  """Defines the gradient of a function to be the straight-through-estimator.

  Specifically, the Jacobian-vector product associated with the function is
  defined to be the identity.

  This causes Jax to effectively ignore this function in the backwards pass.

  Args:
    jax_function: A Jax function that has been decorated with @jax.custom_vjp.
      It is expected to take in and return one positional argument.
  """

  # See
  # https://jax.readthedocs.io/en/latest/notebooks/Custom_derivative_rules_for_Python_code.html
  def ste(primals, tangents):
    return jax_function(primals[0]), tangents[0]

  jax_function.defjvp(ste)


@jax.custom_jvp
def floor_with_gradient(x: jax.Array) -> jax.Array:
  """Floor with Straight-Through-Estimator gradient."""
  return jnp.floor(x)


add_straight_through_estimator(floor_with_gradient)


@jax.custom_jvp
def round_with_gradient(x: jax.Array) -> jax.Array:
  """Round with Straight-Through-Estimator gradient."""
  return jnp.floor(x + jnp.array(0.5))


add_straight_through_estimator(round_with_gradient)


def round_and_clip_to_signed_int(x: jax.Array, *, prec: int, dtype: jnp_dtype,
                                 half_shift: bool) -> jax.Array:
  """Round and clip x to range of signed type with precision 'prec'.

  Requires prec <= 24.

  Args:
    x: The argument to be quantized.
    prec: The target precision for quantization.
    dtype: Desired type of the result.
    half_shift: Uses all available values in a given integer type and makes them
      symmetric by adding 0.5. E.g. instead of to [-1, 0, 1] we will round to
      [-1.5, -0.5, 0.5, 1.5].

  Returns:
    Result.
  """

  # epsilon has to be big enough so that its subtraction in bound computation is
  # not rounded down to zero. It has to be small enough so it has no ML effect.
  # epsilon is necessary for half_shift when prec=1 so that values get floored
  # to -1/0 (not -1/0/1) after clipping to bound.
  epsilon = 2**(-7)
  bound = signed_int_bound(prec=prec, half_shift=half_shift)
  if half_shift:
    bound -= epsilon
    x = jnp.clip(x, a_min=-bound, a_max=bound).astype(dtype)
    x = floor_with_gradient(x) + 0.5
  else:
    # TODO(lew): Use the formula for better gradients. Needs a sweep though.
    # bound = 2**(prec - 1) - 0.5 - epsilon
    x = jnp.clip(x, a_min=-bound, a_max=bound).astype(dtype)
    x = round_with_gradient(x)
  return x


def floor_and_clip_to_unsigned_int(x: jax.Array, *, prec: int,
                                   dtype: jnp_dtype,
                                   half_shift: bool) -> jax.Array:
  """Floor-and-clip x to range of unsigned type with precision 'prec'.

  Requires prec <= 24.

  Args:
    x: The argument to be quantized.
    prec: The target precision for quantization.
    dtype: Desired type of the result.
    half_shift: Needs to be false.

  Returns:
    Result.
  """
  assert not half_shift
  x = floor_with_gradient(x)
  # TODO(lew): should be (a_max=2**prec - epsilon) for a better gradient.
  x = jnp.clip(x, a_min=0, a_max=2**prec - 1).astype(dtype)
  return x


def signed_int_bound(prec: int, half_shift: bool) -> float:
  """Computes the bound value for scaling signed input with precision 'prec'."""
  if prec >= 1:
    # bound
    if half_shift:
      return 2**(prec - 1.0)
    else:
      # TODO(lew): run an experiment with 2bit rounding with 2**(prec-1) - 0.5
      # When prec=1 and half_shift is turned off, a zero bound will make the
      # scaling factor zero. This causes division by zero problem when dividing
      # the scaling factor back. So we need to return a nonzero value that
      # makes inputs rounded to zero.
      return 2**(prec - 1.0) - 1.0 if prec > 1 else 0.25
  else:  # prec < 1
    raise ValueError('prec value should be >= 1.')


def max_abs_weights(x: jax.Array,
                    *,
                    axis: Optional[Iterable[int]] = None) -> BoundsT:
  """Computes the maximum of the absolute value of weights along the axis."""
  abs_max_x = jnp.max(jnp.abs(x), axis=axis, keepdims=True)
  return abs_max_x


def unsigned_int_bound(prec: int) -> float:
  """Computes bound value for scaling unsigned input with precision 'prec'."""
  # NOTE: it's computing 2**prec, which is above the largest unsigned
  # value for that prec. This factor is used for scaling input to range
  # [0, 2**prec].
  if prec < 0:  # TODO(shivaniagrawal) or allow only some values: 4, 8 etc.
    raise ValueError('prec value should be >= 0.')

  return 2**prec



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantization calibration methods."""

import abc
from collections.abc import Sequence
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


def ceil_to_po2(scale: jnp.ndarray) -> jnp.ndarray:
  # With floor the biggest value (we are using jnp.max) is in the range of
  # clipping and therefore have a correct gradient.
  scale = 2 ** jnp.floor(jnp.log2(jax.lax.reciprocal(scale)))
  scale = jax.lax.reciprocal(scale)
  return scale


@utils.flax_slots_kw_only_dataclass
class Calibration(abc.ABC):
  """Abstract class for scale and bias calibration."""

  # The dtype of the quantization scale and bias arrays. If not set, the arrays
  # will be in the same dtype as the input.
  dtype: None | jnp.dtype = utils.static_field(default=None)
  # Round up the calibration to power of 2 (po2).
  po2_scale: bool = utils.static_field(default=False)

  @abc.abstractmethod
  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    """Returns the quantizaiton scale and bias for the given input tensor."""
    # NOTE: The scale and bias calculation are handled by the Calibration
    # class because there is not a single order in which they should be
    # calculated. In the case of symmetric quantization, the scale depends on
    # the bias as the bias shifts the symmetric upperbound. In the case of
    # asymmetric quantization, the bias depends on the scale as the scale
    # determines how far the bias should shift the input s.t. the minimum
    # quantized value aligns with the minimum quantization bucket.
    pass

  def init_calibration(self):
    pass


@utils.flax_slots_kw_only_dataclass
class ConstantCalibration(Calibration):
  """Calibration with a constant per-tensor or per-channel value."""

  bound: jnp.ndarray | float
  bias: None | jnp.ndarray | float = None

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    del context
    if isinstance(self.bound, float) and self.bound <= 0.0:
      raise ValueError(f'{self.bound=} should be positive.')
    dtype = self.dtype if self.dtype is not None else x.dtype

    # TODO(yichizh): hardcode bf16 for the scales, subject to quality evaluation
    bound = self.bound
    if jnp.isscalar(bound):
      bound_shape = list(x.shape)
      for ax in shared_axes:
        bound_shape[ax] = 1
      bound = jnp.ones(bound_shape, dtype=x.dtype) * bound
    scale = bound / numerics_.get_quant_bound()
    scale = ceil_to_po2(scale) if self.po2_scale else scale

    if self.bias is None:
      bias = []
    elif jnp.isscalar(self.bias) or isinstance(self.bias, float):
      # floats are scalars, but pytype can't infer that.
      bias = [jnp.full(x.shape, self.bias, x.dtype)]
    else:
      bias = [self.bias.astype(dtype)]
    return [scale.astype(dtype)], bias


@utils.flax_slots_kw_only_dataclass
class AbsMaxCalibration(Calibration):
  """Simple max(abs(x)) calibration.

  Attributes:
    clipping_scale: Set it to something like 0.3, 0.1, 0.03. If clipping_scale <
      1.0, setting IntSymmetric.clip_gradient=True is likely to be important.
  """

  clipping_scale: None | float = None

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    """Calibration.

    Args:
      x: The input tensor.
      shared_axes: Axes that share a calibration bound. For AbsMaxCalibration,
        it should not be None.
      numerics_: An `AqtNumerics` object containing information regarding
        quantization. Used to create the scale and bias arrays.
      context: The quantization context.

    Returns:
      The scale tensor containing the scale values for each group (can
      potentially be a subchannel). Its shape will be the same as `x.shape` but
      with `shared_axes` collapsed to 1. Bias is not supported.
    """
    del context
    msg = (
        'Perhaps you are using DequantMode.THIS_INPUT (fake_quant) and forgot'
        ' to set them.'
    )
    assert shared_axes is not None, msg
    dtype = self.dtype if self.dtype is not None else x.dtype

    # NOTE: If you use a clipping_scale, consider using clip and clip_gradient
    # in int_numerics.IntSymmetric.
    abs_max = jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True)
    # TODO(yichizh): the zero filtering is not needed anymore because inf is
    # filtered when calculating the reciprocal of scaling factor
    abs_max = jnp.where(abs_max == 0.0, jnp.ones_like(abs_max), abs_max)
    bound = abs_max * self.clipping_scale if self.clipping_scale else abs_max

    scale = bound / numerics_.get_quant_bound()
    scale = ceil_to_po2(scale) if self.po2_scale else scale
    return [scale.astype(dtype)], []


@utils.flax_slots_kw_only_dataclass
class AbsMeanCalibration(Calibration):
  """Simple clipping_scale * mean(abs(x) ** p) ** (1 / p) calibration.

  Attributes:
    clipping_scale: If clipping_scale < 1.0, setting
      IntSymmetric.clip_gradient=True is likely to be important.
    p: Set it to 1 for mean of absolute scaling.
  """

  clipping_scale: float
  p: float

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    """Calibration."""
    del context
    assert shared_axes is not None
    dtype = self.dtype if self.dtype is not None else x.dtype

    abs_sum = jnp.sum(jnp.abs(x) ** self.p, axis=shared_axes, keepdims=True)
    count = jnp.sum(x != 0.0, axis=shared_axes, keepdims=True)
    count = jnp.where(count == 0.0, jnp.ones_like(count), count)
    abs_mean = (abs_sum / count) ** (1.0 / self.p)
    abs_mean = abs_mean * self.clipping_scale
    abs_mean = jnp.where(abs_mean == 0.0, jnp.ones_like(abs_mean), abs_mean)

    scale = abs_mean / numerics_.get_quant_bound()
    scale = ceil_to_po2(scale) if self.po2_scale else scale
    return [scale.astype(dtype)], []


@utils.flax_slots_kw_only_dataclass
class SnrBasedAutoCalibration(Calibration):
  """Automatically finds the best clipping scales based on SNR values.

  The best clipping scales are determined by the SNR (signal-to-noise ratio)
  values of the quantized tensor. The SNR is calculated by the following
  formula:
    SNR = log(1 + signal / noise)
  where signal = sum(x ** 2) and noise = sum(err ** 2).
  An SNR value is calculated for each clipping scale per subchannel group.
  Clipping scales that produce the highest SNR value for each subchannel group
  are selected and used to calculate the best quantization scale.

  Attributes:
    auto_clip_search_config: A sequence of clipping scales to use to search for
      the best per-channel quantization scale.
  """

  auto_clip_search_config: utils.AutoScaleSearchConfig

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    """Produces the scale for quantization based on SNR values.

    Args:
      x: The input tensor.
      shared_axes: Axes that each subchannel group is shared across.
      numerics_: An `AqtNumerics` object containing information regarding
        quantization such as target dtype. Also used to actually quantize (round
        and clip) the tensor when calculating the SNR values.
      context: The quantization context.

    Returns:
      The scale tensor containing the scale values for each subchannel group.
      Its shape will be the same as `x.shape` but with `shared_axes` collapsed
      to 1. Biases are not supported.
    """
    dtype = self.dtype if self.dtype is not None else x.dtype

    # Determine the shape of the best_subchannel_clip_scales. There will be one
    # clip scale per subchannel group, so it shape will be the same as
    # `x.shape` but with `shared_axes` collapsed to 1.
    clip_shape = list(x.shape)
    for i in shared_axes:
      clip_shape[i] = 1

    # Default clipping scale of 1.0 (max value). One per subchannel group.
    best_subchannel_clip_scales = jnp.ones(clip_shape, dtype=jnp.float32)

    # Start with the worst possible SNR value of zeros, essentially representing
    # infinite noise. This will be updated as we search through the clip
    # scales.
    max_snr_values = jnp.zeros(clip_shape, dtype=jnp.float32)

    abs_max = jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True)
    abs_max = jnp.where(abs_max == 0.0, jnp.ones_like(abs_max), abs_max)

    # Iteratively search through the clip scales search space, identifying and
    # updating the best SNR and corresponding clip scale for each subchannel
    # group. Essentially it is performing the "find the max value" for each
    # subgroup in O(auto_clip_search_config) time.
    for clip_scale in self.auto_clip_search_config:
      # Replace the new highest SNR values and corresponding clip scales
      # after evaluating for `clip`.
      best_subchannel_clip_scales, max_snr_values = (
          self._update_best_clip_scales_and_max_snr(
              best_subchannel_clip_scales,
              max_snr_values,
              clip_scale,
              x,
              abs_max,
              shared_axes,
              numerics_,
              context,
          )
      )

    # TODO(b/339746869): Generate a simple report for the clip distribution.
    bound = abs_max * best_subchannel_clip_scales
    scale = bound / numerics_.get_quant_bound()
    scale = ceil_to_po2(scale) if self.po2_scale else scale
    return [scale.astype(dtype)], []

  def _update_best_clip_scales_and_max_snr(
      self,
      current_clip_scales: jnp.ndarray,
      current_snr_values: jnp.ndarray,
      clip_scale: float,
      x: jnp.ndarray,
      abs_max: jnp.ndarray,
      shared_axes: Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: utils.Context,
  ) -> tuple[jnp.ndarray, jnp.ndarray]:
    """Updates the best clip scales and max SNR values given a `clip_scale`.

    Given a `clip_scale`, this function calculates the SNR value for each
    subchannel group. It then identifies the subchannel groups that have higher
    SNR values than `current_snr_values` and updates the best clip scales and
    max SNR values for those groups.

    `current_clip_scales`, `current_snr_values`, and `abs_max` are expected to
    have the same shape, which is the same as `x.shape` but with `shared_axes`
    collapsed to 1.

    `x`, `abs_max`, `shared_axes`, and `context` are only required to calculate
    the SNR values for each subchannel group.

    Args:
      current_clip_scales: The current best clip scales for each subchannel
        group.
      current_snr_values: The current best SNR values for each subchannel group.
      clip_scale: The clip scale to be evaluated.
      x: The input tensor.
      abs_max: The absolute max value for each subchannel group.
      shared_axes: Axes that each subchannel group is shared across.
      numerics_: An `AqtNumerics` object containing information regarding
        quantization such as target dtype. Also used to actually quantize (round
        and clip) the tensor when calculating the SNR values.
      context: The quantization context.

    Returns:
      The (updated best clip scales, updated best SNR values) tuple.
    """
    # Note that all subchannel groups are clipped by the same candidate clip
    # scale.
    clipped_abs_max = abs_max * clip_scale
    clipped_abs_max = clipped_abs_max.astype(x.dtype)

    snr_values = self._calculate_snr(
        x, clipped_abs_max, shared_axes, numerics_, context
    )

    # Update the best clipping scales and SNR values for subchannel groups that
    # have higher SNR values.
    updated_clip_scales = jnp.where(
        snr_values > current_snr_values,
        clip_scale,
        current_clip_scales,
    )

    updated_snr_values = jnp.maximum(snr_values, current_snr_values)
    return updated_clip_scales, updated_snr_values

  def _calculate_snr(
      self,
      x: jnp.ndarray,
      bound: jnp.ndarray,
      shared_axes: Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: utils.Context,
  ) -> jnp.ndarray:
    """Calculates the quantization signal-to-noise ratio (SNR) for the given bound.

    Signal-to-noise = log(1 + signal / noise)
    where noise = sum(err ** 2) and err = x - dequantized_tensor.

    Args:
      x: The input tensor.
      bound: The bound value (scaled absolute max) for each subchannel group.
        Its shape will be the same as `x.shape` but with `shared_axes` collapsed
        to 1.
      shared_axes: Axes that each subchannel group is shared across. SNR values
        will be calculated for each dimension in `x.shape` except the shared
        axes.
      numerics_: An `AqtNumerics` object containing information regarding
        quantization such as target dtype. Also used to actually quantize (round
        and clip) the tensor when calculating the SNR values.
      context: The quantization context.

    Returns:
      The SNR tensor containing the SNR values for each subchannel group. Its
      shape will be the same as `x.shape` but with `shared_axes` collapsed to 1.
    """
    scale = bound / numerics_.get_quant_bound()
    scale = scale.astype(self.dtype if self.dtype is not None else x.dtype)

    qt = aqt_tensor.QTensor(
        qvalue=None,
        scale=[scale],
        scale_t=None,
        bias=[],
        dequant_dtype=x.dtype,
    )
    qt = qt.quant(x)

    # This actually quantizes the tensor (clips, rounds, etc).
    quantized_tensor, _ = numerics_.vjp_fwd(qt.qvalue, context)
    qt.qvalue = quantized_tensor

    dequantized_tensor = qt.dequant()
    err = x - dequantized_tensor

    noise = jnp.sum(err**2, axis=shared_axes, keepdims=True)
    signal = jnp.sum(x**2, axis=shared_axes, keepdims=True)
    snr = jnp.log(1 + signal / noise)

    return snr



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""QTensor states holder for Pallas."""

from __future__ import annotations

from typing import Generic, TypeVar
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import transpose
from aqt.jax.v2 import utils

import flax.struct
import jax  # pylint: disable=unused-import
from jax.experimental import pallas as pl
import jax.numpy as jnp
import numpy as np


T = TypeVar('T')
Array = jnp.ndarray
QTensor = aqt_tensor.QTensor

transpose = transpose.transpose


def _called_within_pallas_kernel(func):
  """Decorator to indicate that the function is called within Pallas kernel."""
  return func


@utils.flax_slots_kw_only_dataclass
class TransposedTensor(Generic[T]):
  """Transposed tensor."""

  transposed_tensor: T
  permute_axes: list[utils.AxisIdx] = flax.struct.field(
      default=None, pytree_node=False
  )

  @property
  @_called_within_pallas_kernel
  def untransposed(self):
    return transpose(
        self.transposed_tensor[...], list(np.argsort(self.permute_axes))
    )


def _count_less(list_, item):
  """Count the number of elements in list_ that are less than item."""
  return len([i for i in list_ if i < item])


def transpose_tensor_for_memory_saving(
    s: jax.Array, block_spec: pl.BlockSpec
) -> (
    tuple[jax.Array, pl.BlockSpec]
    | tuple[TransposedTensor[jax.Array], TransposedTensor[pl.BlockSpec]]
):
  """Transpose given tensor s for memory saving."""

  # If the size of the last dimension is less than 128, padding will be
  # added to make up the difference, which results in wasted memory space.
  # To prevent this, if the second minor most dimension is greater than 1,
  # a transpose between the most minor and second minor most dimensions
  # will be performed.
  transpose_last_two_axes = (s.shape[-1] == 1 and s.shape[-2] > 1)
  if not transpose_last_two_axes:
    return s, block_spec

  permute_axes = list(range(s.ndim))
  permute_axes[-1], permute_axes[-2] = permute_axes[-2], permute_axes[-1]
  transposed_tensor = transpose(s, permute_axes)

  transposed_block_shape = tuple(
      block_spec.block_shape[i] for i in permute_axes
  )

  def transposed_index_map(
      *args,
      permute_axes=tuple(permute_axes),
  ):
    index = list(block_spec.index_map(*args))
    return tuple(index[i] for i in permute_axes)

  # if there are None in block_shape, it means that those axes are
  # reduced when fetched inside pallas kernel. Therefore, we should amend
  # permute_axes accordingly.
  dims_to_be_reduced = [
      i for i, size in enumerate(transposed_block_shape) if size is None
  ]
  permute_axes = [
      i - _count_less(dims_to_be_reduced, i)
      for i in permute_axes
      if i not in dims_to_be_reduced
  ]

  transposed_tensor = TransposedTensor(
      transposed_tensor=transposed_tensor, permute_axes=permute_axes
  )
  transposed_block_spec = transposed_tensor.replace(
      transposed_tensor=pl.BlockSpec(
          block_shape=transposed_block_shape,
          index_map=transposed_index_map,
      )
  )
  return transposed_tensor, transposed_block_spec


def make_qtensor_blockspec(
    qtensor: aqt_tensor.QTensor, block_spec: pl.BlockSpec
) -> QTensor:
  """Build a block spec for QTensor.

  Args:
    qtensor: a QTensor.
    block_spec: Block spec of the unquantized tensor.

  Returns:
    A QTensor of block spec.
  """

  assert qtensor.scale_t is None

  def _make_scale_block_spec(scale: jax.Array, block_spec: pl.BlockSpec):
    scale_blk_shape = list(block_spec.block_shape)

    # Find calibration axes, and change scale_block_shape accordingly.
    calibration_axes = [axes for axes, s in enumerate(scale.shape) if s == 1]

    for i in calibration_axes:
      if scale_blk_shape[i] is not None:
        scale_blk_shape[i] = 1

    def scale_index_map(
        *args,
        calibration_axes=calibration_axes,
    ):
      index = list(block_spec.index_map(*args))
      for i in calibration_axes:
        index[i] = 0
      return tuple(index)

    return pl.BlockSpec(
        block_shape=tuple(scale_blk_shape),
        index_map=scale_index_map,
    )

  return QTensor(
      qvalue=block_spec,
      scale=[_make_scale_block_spec(s, block_spec) for s in qtensor.scale],
      scale_t=None,
      bias=[],
      dequant_dtype=qtensor.dequant_dtype,
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AQT calibration module with logic from Transformer Engine."""
from typing import Sequence

from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import flax.linen as nn
import jax
from jax import numpy as jnp

CALIBRATION_STATS = "calibration_stats"


@utils.flax_slots_kw_only_dataclass
class DelayedScalingCalibration(calibration.Calibration, nn.Module):
  """Calibration module with logic from Transformer Engine, utilizing Delayed Scaling."""

  amax_history_length: int = 1024

  def setup(self) -> None:
    # If we use nn.compact, we have to have a dummy call to self.get_bound()
    # during initialization. This is particularly problematic when using this
    # module for backward pass calibration, when get_bound() isn't naturally
    # invoked during the dummy forward pass during init().

    # Using setup() keeps things simpler and avoids the need to make awkward
    # dummy calls to get_bound().
    self.amax_history = self.variable(
        CALIBRATION_STATS,
        "amax_history",
        # pylint: disable-next=protected-access
        lambda: jax._src.core.mutable_array(
            jnp.zeros((self.amax_history_length,))
        ),
    )

    self.bound = self.variable(
        CALIBRATION_STATS,
        "bound",
        # pylint: disable-next=protected-access
        lambda: jax._src.core.mutable_array(jnp.zeros((1,))),
    )

  def get_bound(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      context: None | utils.Context = None,
  ) -> jnp.ndarray:
    del shared_axes
    # Right now we just support per_tensor calibration (i.e. one value).
    # To support per_axis calibration, we would need to be able to change the
    # shape of the mutable arrays. For example, right now amax_history has
    # shape (amax_history_length,). But if we want to support per_axis
    # calibration, we would need the shape of amax_history to be changed to
    # (amax_history_length, scale_shape), where scale_shape is
    # jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True).shape
    #
    # See setup() for why we can't use nn.compact to solve this as is usually
    # done Jax/Flax

    # We default to SERVE mode if context is not provided, since we will
    # be mutating the arrays in place and don't want to do so accidentally
    quant_mode = context.quant_mode if context else utils.QuantMode.SERVE

    bound_mutable_arr = self.bound.value
    amax_history_mutable_arr = self.amax_history.value

    amax_history = amax_history_mutable_arr[:]
    prev_bound = bound_mutable_arr[:]
    amax_from_history = jnp.max(amax_history, axis=0)

    new_bound = self.compute_bound(amax_from_history, prev_bound)
    new_history = self.compute_history(x, amax_history)

    if quant_mode in [utils.QuantMode.TRAIN, utils.QuantMode.CALIBRATE]:
      bound_mutable_arr[:] = new_bound[:]
      amax_history_mutable_arr[:] = new_history[:]
    return new_bound.reshape((1,) * len(x.shape))

  def get_scale_and_bias(
      self,
      x: jnp.ndarray,
      shared_axes: None | Sequence[utils.AxisIdx],
      numerics_: numerics.AqtNumerics,
      context: None | utils.Context = None,
  ) -> tuple[list[jnp.ndarray], list[jnp.ndarray]]:
    dtype = self.dtype if self.dtype is not None else x.dtype
    bound = self.get_bound(x, shared_axes, context)
    scale = bound / numerics_.get_quant_bound()
    scale = calibration.ceil_to_po2(scale) if self.po2_scale else scale
    return [scale.astype(dtype)], []

  def compute_bound(self, amax, prev_bound):
    new_bound = jnp.copy(amax)
    new_bound = jnp.where(amax > 0.0, new_bound, prev_bound)
    new_bound = jnp.where(jnp.isfinite(amax), new_bound, prev_bound)
    return new_bound

  def compute_history(self, x, amax_history):
    amax_update = jnp.max(jnp.abs(x)).astype(amax_history.dtype)
    new_history = (
        jnp.roll(amax_history, shift=-1, axis=0).at[0].set(amax_update)
    )
    return new_history

  def init_calibration(self):
    # We need to "touch" one of these variables to make sure this modules
    # variables are initialized properly once. Else, they get "recreated"
    # on each use.
    self.amax_history  # pylint: disable=pointless-statement



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt_matmul."""

from typing import Iterable

from absl.testing import absltest
from aqt.common import aqt_config
from aqt.jax import aqt_dot_general
from aqt.jax import aqt_ops
from aqt.jax import aqt_tensor
from aqt.test import aqt_matmul_test_base
from flax import linen as nn
import jax
import jax.numpy as jnp

# pylint: disable=g-long-lambda

dot_general = aqt_dot_general.injectable_dot_general


class Matmul(nn.Module):
  config: aqt_config.AqtMatmulConfig
  lhs_shape: Iterable[int]
  rhs_shape: Iterable[int]

  def setup(self):
    self.lhs_quantizer = aqt_tensor.TensorQuantizer(
        list(self.lhs_shape), self.config.lhs)
    self.rhs_quantizer = aqt_tensor.TensorQuantizer(
        list(self.rhs_shape), self.config.rhs)

  def __call__(self, lhs, rhs, train=True):
    return aqt_ops.aqt_matmul(
        lhs,
        rhs,
        dot_general=dot_general(self.lhs_quantizer, self.rhs_quantizer, train))

  def update_lhs(self, x, weight, event_count):
    self.lhs_quantizer.update(x, weight, event_count)

  def update_rhs(self, x, weight, event_count):
    self.rhs_quantizer.update(x, weight, event_count)


class AqtMatmulTest(aqt_matmul_test_base.MatmulTest):

  _matmul_state = {}

  def constant(self, x):
    return jnp.array(x, dtype=jnp.float32)

  def matmul(self, config, lhs_shape, rhs_shape, name="aqt"):
    mm = Matmul(config, lhs_shape, rhs_shape, name=name)
    state = mm.init(
        jax.random.PRNGKey(0), jnp.zeros(lhs_shape), jnp.zeros(rhs_shape))
    self._matmul_state[name] = state

    return mm

  def matmul_apply(self, mm, lhs, rhs, train=True, keep_stats=False):
    event_count = 0
    lhs_sample = jnp.zeros_like(lhs) if keep_stats else lhs
    lhs_weight = jnp.ones_like(lhs) if keep_stats else None
    rhs_sample = jnp.zeros_like(rhs) if keep_stats else rhs
    rhs_weight = jnp.ones_like(rhs) if keep_stats else None

    _, state = mm.apply(
        self._matmul_state[mm.name],
        lhs_sample,
        lhs_weight,
        event_count,
        method=mm.update_lhs,
        mutable=True)
    _, state = mm.apply(
        state,
        rhs_sample,
        rhs_weight,
        event_count,
        method=mm.update_rhs,
        mutable=True)
    self._matmul_state[mm.name] = state

    return mm.apply(self._matmul_state[mm.name], lhs, rhs, train)

  def matmul_unquantized(self, lhs, rhs):
    lhs = self.constant(lhs)
    rhs = self.constant(rhs)

    return jnp.matmul(lhs, rhs)

  def gradients(self, fwd_func, x, w, reduce_sum=False):
    if reduce_sum:
      fwd_func = lambda x, w: jnp.sum(fwd_func(x, w)**2)
    y, vjp_fn = jax.vjp(fwd_func, x, w)
    return vjp_fn(jnp.ones(y.shape))[0]


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is forked from third_party/py/flax/struct.py to optionally make
# `flax.struct.dataclass` mutable. That is useful in our codebase since
# we construct a 'base model configuration' as an HParams dataclass and then
# want to mutate it in the axes handlers for running experiment sweeps.
#
# We default unsafe_hash to True in our fork since the Jax compiler caches
# static parameters based on their hash. The hash of a mutable dataclass is
# by default its object id, and so mutating its fields won't change its
# hash and will cause Jax to incorrectly use a cached version of JITed
# functions that don't reflect new mutations. Setting unsafe_hash causes the
# hash to instead be based on the value of the dataclass, and so Jax will
# correctly recompile when a field has changed.
"""Utilities for defining custom classes that can be used with jax transformations."""

import dataclasses
from flax import serialization
import jax


def dataclass(clz=None, *, frozen=False, unsafe_hash=True):
  """Create a class which can be passed to functional transformations.

  Jax transformations such as `jax.jit` and `jax.grad` require objects that are
  immutable and can be mapped over using the `jax.tree_util` methods.
  The `dataclass` decorator makes it easy to define custom classes that can be
  passed safely to Jax. For example::
    from flax import struct
    @struct.dataclass
    class Model():
      params: Any
      # use pytree_node=False to indicate an attribute should not be touched
      # by Jax transformations.
      apply_fn: FunctionType = struct.field(pytree_node=False)
      def __apply__(self, *args):
        return self.apply_fn(*args)
    model = Model(params, apply_fn)
    model.params = params_b  # Model is immutable. This will raise an error.
    model_b = model.replace(params=params_b)  # Use the replace method instead.
    # This class can now be used safely in Jax to for example to compute
    # gradients w.r.t. the parameters.
    model = Model(params, apply_fn)
    model_grad = jax.grad(some_loss_fn)(model)
  Args:
    clz: the class that will be transformed by the decorator.
    frozen: whether to freeze the dataclass (default=False). WARNING: mutations
      that occur to an unfrozen dataclass within a function that has been
      transformed by Jax (including Flax modules) are not visible outside of
      that function.
    unsafe_hash: see help for corresponding argument in dataclasses.dataclass.

  Returns:
    The new class.
  """

  def wrapped(clz):
    data_clz = dataclasses.dataclass(
        frozen=frozen, unsafe_hash=unsafe_hash)(
            clz)
    meta_fields = []
    data_fields = []
    for name, field_info in data_clz.__dataclass_fields__.items():
      is_pytree_node = field_info.metadata.get('pytree_node', True)
      if is_pytree_node:
        data_fields.append(name)
      else:
        meta_fields.append(name)

    def replace(self, **updates):
      """"Returns a new object replacing the specified fields with new values."""
      return dataclasses.replace(self, **updates)

    data_clz.replace = replace

    def iterate_clz(x):
      meta = tuple(getattr(x, name) for name in meta_fields)
      data = tuple(getattr(x, name) for name in data_fields)
      return data, meta

    def clz_from_iterable(meta, data):
      meta_args = tuple(zip(meta_fields, meta))
      data_args = tuple(zip(data_fields, data))
      kwargs = dict(meta_args + data_args)
      return data_clz(**kwargs)

    jax.tree_util.register_pytree_node(data_clz, iterate_clz, clz_from_iterable)

    def to_state_dict(x):
      state_dict = {
          name: serialization.to_state_dict(getattr(x, name))
          for name in data_fields
      }
      return state_dict

    def from_state_dict(x, state):
      """Restore the state of a data class."""
      state = state.copy()  # copy the state so we can pop the restored fields.
      updates = {}
      for name in data_fields:
        if name not in state:
          raise ValueError(f'Missing field {name} in state dict while restoring'
                           f' an instance of {clz.__name__}')
        value = getattr(x, name)
        value_state = state.pop(name)
        updates[name] = serialization.from_state_dict(value, value_state)
      if state:
        names = ','.join(state.keys())
        raise ValueError(f'Unknown field(s) "{names}" in state dict while'
                         f' restoring an instance of {clz.__name__}')
      return x.replace(**updates)

    serialization.register_serialization_state(data_clz, to_state_dict,
                                               from_state_dict)

    return data_clz

  # If this is called as @dataclass(frozen=...), clz is will be None and
  # and we return a function that will be later called with clz.
  if clz is None:
    return wrapped
  # Otherwise, this was called directly as @dataclass so we apply the
  # wrapper to the clz immediately.
  else:
    return wrapped(clz)


def field(pytree_node=True, **kwargs):
  return dataclasses.field(metadata={'pytree_node': pytree_node}, **kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods to create, save and interact with hlo proto objects."""
import os.path
import re
from typing import Any, Callable, Sequence, Text, Tuple, Union
from flax import linen as nn
import jax
import jax.numpy as jnp


# pylint: disable=g-direct-tensorflow-import
from tensorflow.compiler.xla.service import hlo_pb2
# pylint: enable=g-direct-tensorflow-import


def load_hlo_proto_from_jax_fn(fn: Callable[..., Any], *fn_args: Any,
                               **fn_kwargs: Any) -> hlo_pb2.HloModuleProto:
  """Loads HLO proto object from jax function.

  Args:
    fn: a jax function.
    *fn_args: Arguments to fn.
    **fn_kwargs: Keyword arguments to fn.

  Returns:
    An HloModuleProto object.

  """
  computation = (
      jax.jit(fn).lower(*fn_args, **fn_kwargs).compiler_ir(dialect='hlo')
  )
  serialized_hlo = computation.as_serialized_hlo_module_proto()
  hlo_module_proto = hlo_pb2.HloModuleProto.FromString(serialized_hlo)
  return hlo_module_proto


def load_hlo_proto_from_model(
    model: nn.Module,  #
    state: Any,
    input_shapes: Sequence[Tuple[int, ...]],
    **kwargs) -> hlo_pb2.HloModuleProto:
  """Loads HLO proto object from flax model.

  Args:
    model: a flax model.
    state: for model with stateful context, state of the model.
    input_shapes: Sequence of shapes of model input tensor(s). Order of input
      shapes must correspond to order in which model accepts the inputs.
    **kwargs: Addition arguments forwarded to model's `apply` function.

  Returns:
    An HloModuleProto object.

  """

  ones_shape_list = [jnp.ones(shape) for shape in input_shapes]
  return load_hlo_proto_from_jax_fn(lambda *x: model.apply(state, *x, **kwargs),
                                    *ones_shape_list)


# TODO(shivaniagrawal): use strict pytype here.
def output_hlo(computation: Any, file_path: str):
  """Saves HLO for the given xla computation to given file path.

  The file format is determined by the file_path extension, i.e. for .txt file
  saves the hlo as text, and for .pb file save the hlo module proto binary.

  Args:
    computation: wrapped XlaComputation
    file_path: file path to write the HLO to.
  """
  hlo_module_proto_str = computation.as_serialized_hlo_module_proto()
  hlo_txt = computation.as_hlo_text()
  output_hlo_to_file(hlo_module_proto_str, hlo_txt, file_path)




def count_ops_in_hlo_proto(hlo_proto: hlo_pb2.HloModuleProto,
                           ops_regex: str) -> int:
  """Counts specific ops in hlo proto, whose names match the provided pattern.

  Args:
    hlo_proto: an HloModuleProto object.
    ops_regex: a string regex to filter ops with matching names.

  Returns:
    Count of matching ops.

  """
  ops_count = 0
  for computation in hlo_proto.computations:
    for instr in computation.instructions:
      if re.match(ops_regex, instr.name) is not None:
        ops_count += 1
  return ops_count



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flax layer for AQT injection."""

# pylint: disable=unnecessary-lambda
# pylint: disable=g-importing-member
import copy
import enum
import functools
from typing import Callable, Iterable, Sequence
from aqt.jax.v2 import aqt_conv_general
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import transpose
from aqt.jax.v2 import utils
from aqt.jax.v2.flax import aqt_flax_dg_core
from aqt.jax.v2.flax import freezer as general_freezer
from aqt.jax.v2.utils import QuantMode
import flax.core.meta as nn_meta
import flax.linen as nn
import jax
import jax.numpy as jnp


NoShardingAxes = Sequence[utils.AxisIdx]
AxisMetadataWrapper = Callable[
    [jnp.ndarray, None | tiled_dot_general.AqtTileMap, NoShardingAxes],
    nn_meta.AxisMetadata,
]
DotGeneralTilingFn = Callable[
    [jnp.ndarray, jnp.ndarray, jax.lax.DotDimensionNumbers],
    tiled_dot_general.Cfg
]
EinsumTilingFn = Callable[
    [tiled_dot_general.EinsumEqnLetter, jnp.ndarray, jnp.ndarray],
    tiled_dot_general.Cfg
]


_QUANT_TO_FREEZER_MODE = {
    QuantMode.TRAIN: general_freezer.FreezerMode.NONE,
    QuantMode.CALIBRATE: general_freezer.FreezerMode.NONE,
    QuantMode.CONVERT: general_freezer.FreezerMode.WRITE,
    QuantMode.SERVE: general_freezer.FreezerMode.READ,
}


def _freezer_qtensor_init_wrapper(
    qt: aqt_tensor.QTensor,
    contracting_axis: Sequence[utils.AxisIdx],
    axis_metadata_wrapper: None | AxisMetadataWrapper,
    tile_map: tiled_dot_general.AqtTileMap,
):
  """QTensor initialization wrapper function for the new freezer."""
  # Copy to avoid mutating the input QTensor here or downstream (regardless of
  # axis_metadata_wrapper's value).
  qt = copy.deepcopy(qt)
  if axis_metadata_wrapper is None:
    return qt

  scale_non_shard_axis_all = list(range(qt.ndim))
  scale_non_shard_axis_contracting = list(contracting_axis)

  def _get_singleton_axes(x: jnp.ndarray) -> list[utils.AxisIdx]:
    return [axis for axis, dim in enumerate(x.shape) if dim == 1]

  qt.qvalue = axis_metadata_wrapper(qt.qvalue, tile_map, [])
  qt.scale = jax.tree.map(
      lambda x: axis_metadata_wrapper(
          x, tile_map, scale_non_shard_axis_contracting
      ),
      qt.scale,
  )
  # Passing scale_non_shard_axis_contracting would be incorrect due to
  # scale transposition. scale_t is being removed from QTensor anyway
  # so we just pass scale_non_shard_axis_all.
  qt.scale_t = jax.tree.map(
      lambda x: axis_metadata_wrapper(x, tile_map, scale_non_shard_axis_all),
      qt.scale_t,
  )
  # Set the non-sharding axes for bias to the singleton dimensions.
  qt.bias = jax.tree.map(
      lambda x: axis_metadata_wrapper(x, tile_map, _get_singleton_axes(x)),
      qt.bias,
  )
  return qt


class FreezerMode(enum.Enum):
  NONE = 1
  CALIBRATION = 2
  CALIBRATION_AND_VALUE = 3


class Freezer(nn.Module):
  """Identity function that can freeze its input.

  On default it is an identity function that saves the input in a variable.
  In 'quant_mode=QuantMode.Serve' mode, ignores the input and returns the frozen
  value. It is useful to implement 'constant folding' and put quantized weights
  and scales in the checkpoint for serving. Specifically:

  self.get() returns None when quant_mode=TRAIN or CONVERT, returns variable
  when quant_mode=SERVE.
  self.set() does nothing when quant_mode=TRAIN or SERVE, creates and stores
  quantized tensor when quant_mode=CONVERT.
  """

  quant_collection: str
  quant_mode: QuantMode
  q_shape: Iterable[int]
  q_dtype: jnp.dtype
  q_init: nn.initializers.Initializer
  s_shape: Iterable[int]
  s_init: nn.initializers.Initializer

  def setup(self):
    mode = self.quant_mode
    if mode == QuantMode.SERVE or mode == QuantMode.CONVERT:
      collection = self.quant_collection
      q_init = self.q_init
      q_shape = self.q_shape
      q_dtype = self.q_dtype
      s_init = self.s_init
      s_shape = self.s_shape
      # TODO(lew): Store whole QTensor?
      # We could have created one self.variable whose value is a QTensor,
      # but we are unsure how this would complicate the init function,
      # which could potentially be used by adding metadata such as
      # sharding axises, etc.
      self.qvalue = self.variable(collection, 'value', q_init, q_shape, q_dtype)
      self.scale_t = self.variable(collection, 'scale', s_init, s_shape)

  def get(self) -> None | aqt_tensor.QTensor:
    if self.quant_mode == QuantMode.TRAIN:
      return None
    elif self.quant_mode == QuantMode.CALIBRATE:
      return None
    elif self.quant_mode == QuantMode.CONVERT:
      return None
    elif self.quant_mode == QuantMode.SERVE:
      return aqt_tensor.QTensor(
          qvalue=self.qvalue.value,
          scale=None,
          scale_t=[self.scale_t.value],
          bias=[],
          # TODO(lew): Ideal solution: To find out this dequant_dtype one should
          # use the dtype of inputs of the quant function. We should store it as
          # a dtype of small-sized scale tensor.
          dequant_dtype=self.scale_t.value.dtype,
      )
    else:
      assert False, 'Unknown quant mode.'

  def set(self, inputs: aqt_tensor.QTensor) -> None:
    # TODO(b/325626080): Uncommenting the assert will fail.
    # assert inputs.qvalue.dtype == self.q_dtype, (
    #     f'Freezer got a QTensor of type {inputs.qvalue.dtype} but expected'
    #     f' {self.q_dtype}.'
    # )
    if inputs.bias:
      raise NotImplementedError(
          'Quantization biases are not supported in AQT Flax Legacy Freezer.'
      )
    if self.quant_mode == QuantMode.TRAIN:
      pass
    elif self.quant_mode == QuantMode.CALIBRATE:
      pass
    elif self.quant_mode == QuantMode.CONVERT:
      self.qvalue.value = inputs.qvalue
      assert inputs.scale_t is not None and len(inputs.scale_t) == 1
      self.scale_t.value = inputs.scale_t[0]
    elif self.quant_mode == QuantMode.SERVE:
      # TODO(lew): Optionally compare stored and served value.
      pass
    else:
      assert False, 'Unknown quant mode.'
    return None


def _maybe_recover_scale_from_scale_t(
    qt: None | aqt_tensor.QTensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    is_rhs: bool,
    lhs_shape: Sequence[int],
    rhs_shape: Sequence[int],
) -> aqt_tensor.QTensor:
  """Recovers scale from scale_t if necessary."""
  if qt is None or qt.scale is not None or qt.scale_t is None:
    return qt

  transpose_fn = transpose.lhs_recover_scale_from_scale_t
  if is_rhs:
    transpose_fn = transpose.rhs_recover_scale_from_scale_t
  qt.scale = [
      transpose_fn(scale_t, dimension_numbers, lhs_shape, rhs_shape)
      for scale_t in qt.scale_t
  ]
  qt.scale_t = None
  return qt


def _populate_scale_t(
    qt: aqt_tensor.QTensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    is_rhs: bool,
    lhs_shape: Sequence[int],
    rhs_shape: Sequence[int],
) -> aqt_tensor.QTensor:
  """Populates scale_t from scale."""
  if qt.scale is None:
    return qt

  transpose_fn = transpose.lhs_scale_transpose_to_output
  if is_rhs:
    transpose_fn = transpose.rhs_scale_transpose_to_output
  qt.scale_t = [
      transpose_fn(scale, dimension_numbers, lhs_shape, rhs_shape)
      for scale in qt.scale
  ]
  return qt


class AqtDotGeneral(nn.Module):
  """A layer that can be injected into flax.nn.Dense, etc."""

  cfg: None | aqt_dot_general.DotGeneral = None
  prng_name: None | str = 'params'

  # TODO(lew): split out separate class for each side.
  # Quant mode determines whether flax variables are created to store quantized
  # inputs. Refer to the Freezer doc str to see variable creation in each mode.
  lhs_quant_mode: QuantMode = QuantMode.TRAIN
  # apply_quant_mode determines if using Freezer in cfg.get/set_tensor
  lhs_apply_quant_mode: bool = True
  lhs_var_name: str = 'qlhs'
  lhs_qtensor: None | aqt_tensor.QTensor = None

  rhs_quant_mode: QuantMode = QuantMode.TRAIN
  rhs_apply_quant_mode: bool = True
  rhs_var_name: str = 'qrhs'
  rhs_qtensor: None | aqt_tensor.QTensor = None

  # Variables only for the legacy Freezer.
  lhs_init: nn.initializers.Initializer = jnp.zeros
  lhs_scale_init: nn.initializers.Initializer = jnp.zeros

  rhs_init: nn.initializers.Initializer = jnp.zeros
  rhs_scale_init: nn.initializers.Initializer = jnp.zeros

  # Variables only for the new Freezer.
  lhs_axis_metadata_wrapper: None | AxisMetadataWrapper = None
  rhs_axis_metadata_wrapper: None | AxisMetadataWrapper = None

  # Freeze mode. Set as FreezerMode.CALIBRATION to store only scales; set as
  # CALIBRATION_AND_VALUE to store both scales and quantized values.
  lhs_freeze_mode: FreezerMode = FreezerMode.NONE
  rhs_freeze_mode: FreezerMode = FreezerMode.NONE

  # If you want use 'params' make sure that there is another mechanism to hide
  # these variables from the optimizer.
  quant_collection: str = 'aqt'

  # Tiling configs. tilling_fn is valid only when tiling_cfg is None.
  tiling_cfg: None | tiled_dot_general.Cfg = None
  tiling_fn: None | DotGeneralTilingFn = None

  # If set to True, use the current Freezer. Otherwise, use the new
  # QuantFreezer.
  use_legacy_freezer: bool = True

  def make_aqt_dg(
      self,
      lhs_shape,
      rhs_shape,
      dimension_numbers: tuple[Iterable[int], Iterable[int]],
      lhs_tile_map: None | tiled_dot_general.AqtTileMap = None,
      rhs_tile_map: None | tiled_dot_general.AqtTileMap = None,
  ):
    if self.cfg is None:
      return jax.lax.dot_general

    cfg = copy.deepcopy(self.cfg)
    lhs_scale_shape = list(lhs_shape)
    rhs_scale_shape = list(rhs_shape)
    (contr, _) = dimension_numbers
    for li, ri in zip(*contr):
      lhs_scale_shape[li] = 1
      rhs_scale_shape[ri] = 1
    lhs_scale = transpose.lhs_scale_transpose_to_output(
        jnp.zeros(lhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
    )
    assert lhs_scale is not None
    lhs_scale_shape = lhs_scale.shape
    rhs_scale = transpose.rhs_scale_transpose_to_output(
        jnp.zeros(rhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
    )
    assert rhs_scale is not None
    rhs_scale_shape = rhs_scale.shape
    rhs_qm = self.rhs_quant_mode
    lhs_qm = self.lhs_quant_mode

    assert isinstance(
        cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
    )
    lhs_q_dtype = cfg.fwd.dg_quantizer.lhs.numerics.get_dtype()
    rhs_q_dtype = cfg.fwd.dg_quantizer.rhs.numerics.get_dtype()

    if self.use_legacy_freezer:
      lhs_freezer = Freezer(
          name=self.lhs_var_name,
          quant_mode=lhs_qm,
          q_shape=lhs_shape,
          q_dtype=lhs_q_dtype,
          q_init=self.lhs_init,
          s_shape=lhs_scale_shape,
          s_init=self.lhs_scale_init,
          quant_collection=self.quant_collection,
      )

      rhs_freezer = Freezer(
          name=self.rhs_var_name,
          quant_mode=rhs_qm,
          q_shape=rhs_shape,
          q_dtype=rhs_q_dtype,
          q_init=self.rhs_init,
          s_shape=rhs_scale_shape,
          s_init=self.rhs_scale_init,
          quant_collection=self.quant_collection,
      )
    else:
      lhs_ca, rhs_ca = contr
      lhs_init_wrapper = functools.partial(
          _freezer_qtensor_init_wrapper,
          contracting_axis=lhs_ca,
          axis_metadata_wrapper=self.lhs_axis_metadata_wrapper,
          tile_map=lhs_tile_map,
      )
      rhs_init_wrapper = functools.partial(
          _freezer_qtensor_init_wrapper,
          contracting_axis=rhs_ca,
          axis_metadata_wrapper=self.rhs_axis_metadata_wrapper,
          tile_map=rhs_tile_map,
      )

      lhs_freezer = general_freezer.Freezer(
          name=self.lhs_var_name,
          mode=_QUANT_TO_FREEZER_MODE[lhs_qm],
          collection=self.quant_collection,
          axis_metadata_wrapper=lhs_init_wrapper,
      )

      rhs_freezer = general_freezer.Freezer(
          name=self.rhs_var_name,
          mode=_QUANT_TO_FREEZER_MODE[rhs_qm],
          collection=self.quant_collection,
          axis_metadata_wrapper=rhs_init_wrapper,
      )

    prng_name = self.prng_name
    key = self.make_rng(prng_name) if prng_name is not None else None
    cfg = config.set_context(
        cfg,
        key,
        train_step=None,
        lhs_quant_mode=self.lhs_quant_mode,
        rhs_quant_mode=self.rhs_quant_mode,
    )

    def ret_dg(
        lhs,
        rhs,
        dimension_numbers,
        precision=None,
        preferred_element_type=None,
    ):
      del preferred_element_type
      assert (
          precision is None
      ), f'Precision {precision} requested together with quantization.'

      # TODO(yichizh): asserting xhs dtype only when apply_quant_mode=False
      # and cfg.get_qtensor() is None
      msg = 'AQT is not yet optimized to accept quantized types directly. '
      msg += f'lhs.dtype: {lhs.dtype}, rhs.dtype: {rhs.dtype}'
      assert lhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
      assert rhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg

      cfg.assert_config_validity()

      # Getter
      lhs_apply_quant_mode = self.lhs_apply_quant_mode
      rhs_apply_quant_mode = self.rhs_apply_quant_mode
      lhs_qt = lhs_freezer.get() if lhs_apply_quant_mode else self.lhs_qtensor
      rhs_qt = rhs_freezer.get() if rhs_apply_quant_mode else self.rhs_qtensor

      # Recover scale from scale_t, if necessary.
      # The quantized tensor loaded from the legacy freezer has only scale_t.
      lhs_qt = _maybe_recover_scale_from_scale_t(
          lhs_qt, dimension_numbers, False, lhs_shape, rhs_shape
      )
      rhs_qt = _maybe_recover_scale_from_scale_t(
          rhs_qt, dimension_numbers, True, lhs_shape, rhs_shape
      )

      cfg.apply_custom_vjp_on_jax = False
      out, (out_lhs_qt, out_rhs_qt) = aqt_flax_dg_core.dg_core_flax_lifted(
          lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, self, cfg
      )

      # Remove qvalue of the activation to not to store it in Freezer.
      match self.lhs_freeze_mode:
        case FreezerMode.NONE:
          if self.lhs_apply_quant_mode and self.lhs_quant_mode in {
              QuantMode.CONVERT,
              QuantMode.SERVE,
          }:
            raise ValueError('Freezer is used with Freezer mode NONE.')
        case FreezerMode.CALIBRATION:
          out_lhs_qt = out_lhs_qt.without_qvalue()
        case FreezerMode.CALIBRATION_AND_VALUE:
          pass
        case _:
          raise ValueError('Unknown freeze mode: %s' % self.lhs_freeze_mode)

      match self.rhs_freeze_mode:
        case FreezerMode.NONE:
          if self.rhs_apply_quant_mode and self.rhs_quant_mode in {
              QuantMode.CONVERT,
              QuantMode.SERVE,
          }:
            raise ValueError('Freezer is used with Freezer mode NONE.')
        case FreezerMode.CALIBRATION:
          out_rhs_qt = out_rhs_qt.without_qvalue()
        case FreezerMode.CALIBRATION_AND_VALUE:
          pass
        case _:
          raise ValueError('Unknown freeze mode: %s' % self.rhs_freeze_mode)

      # Setter
      calib_contracting_axis = aqt_dot_general.CalibrationMode.CONTRACTING_AXIS
      if self.use_legacy_freezer:
        # We need to populate the stored QTensor with scale_t.
        if cfg.fwd.lhs.calibration_mode == calib_contracting_axis:
          out_lhs_qt = _populate_scale_t(
              out_lhs_qt, dimension_numbers, False, lhs_shape, rhs_shape
          )
        if cfg.fwd.rhs.calibration_mode == calib_contracting_axis:
          out_rhs_qt = _populate_scale_t(
              out_rhs_qt, dimension_numbers, True, lhs_shape, rhs_shape
          )

      if self.lhs_apply_quant_mode:
        lhs_freezer.set(out_lhs_qt)
      if self.rhs_apply_quant_mode:
        rhs_freezer.set(out_rhs_qt)

      return out

    return ret_dg

  @nn.compact
  def __call__(
      self,
      lhs,
      rhs,
      dimension_numbers,
      precision,
      preferred_element_type=None,
  ):
    tiling_cfg = self.tiling_cfg
    if tiling_cfg is None and self.tiling_fn is not None:
      tiling_cfg = self.tiling_fn(lhs, rhs, dimension_numbers)

    if tiling_cfg is not None:
      xlhs, xrhs = tiled_dot_general.generate_tiling_states_for_dot_general(
          tiling_cfg, lhs, rhs, dimension_numbers
      )
      # Extract tiled input shapes and dimension numbers from jaxpr
      def dummy_tiled_dg(lhs_in, rhs_in):
        return tiled_dot_general.tiled_dot_general(
            tiling_cfg, lhs_in, rhs_in, dimension_numbers
        )

      tiled_dg_jaxpr = jax.make_jaxpr(dummy_tiled_dg)(lhs, rhs)
      dg_eqn = [eqn for eqn in tiled_dg_jaxpr.eqns if 'dot_general' in str(eqn)]
      assert len(dg_eqn) == 1, 'Multiple dg calls are found in tiled dg jaxpr.'
      lhs_invar, rhs_invar = dg_eqn[0].invars
      tiled_lhs_shape = lhs_invar.aval.shape
      tiled_rhs_shape = rhs_invar.aval.shape
      tiled_dimension_numbers = dg_eqn[0].params['dimension_numbers']
      # Use tiled input shapes and dimension numbers to create aqt_dg that
      # will be injected to tiled_dot_general
      aqt_dg = self.make_aqt_dg(
          tiled_lhs_shape,
          tiled_rhs_shape,
          tiled_dimension_numbers,
          lhs_tile_map=xlhs.tile_map,
          rhs_tile_map=xrhs.tile_map,
      )
      # We integrate tiling here and not on Jax level, so that the Freezers
      # observe tiled shapes.
      ret_dg = functools.partial(
          tiled_dot_general.tiled_dot_general,
          tiling_cfg,
          dot_general=aqt_dg,
      )
    else:
      ret_dg = self.make_aqt_dg(lhs.shape, rhs.shape, dimension_numbers)
    return ret_dg(
        lhs,
        rhs,
        dimension_numbers,
        precision,
        preferred_element_type,
    )


class AqtEinsum(nn.Module):
  """Quantized Einsum class for model injection."""

  cfg: None | config.DotGeneral = None
  prng_name: None | str = 'params'

  # TODO(lew): split out separate class for each side.
  lhs_quant_mode: QuantMode = QuantMode.TRAIN
  lhs_var_name: str = 'qlhs'

  rhs_quant_mode: QuantMode = QuantMode.TRAIN
  rhs_var_name: str = 'qrhs'

  # Variables only for the legacy Freezer.
  lhs_init: nn.initializers.Initializer = jnp.zeros
  lhs_scale_init: nn.initializers.Initializer = jnp.zeros

  rhs_init: nn.initializers.Initializer = jnp.zeros
  rhs_scale_init: nn.initializers.Initializer = jnp.zeros

  # Variables only for the new Freezer.
  lhs_axis_metadata_wrapper: None | AxisMetadataWrapper = None
  rhs_axis_metadata_wrapper: None | AxisMetadataWrapper = None

  # Freeze mode. Set as FreezerMode.CALIBRATION to store only scales; set as
  # CALIBRATION_AND_VALUE to store both scales and quantized values.
  lhs_freeze_mode: FreezerMode = FreezerMode.NONE
  rhs_freeze_mode: FreezerMode = FreezerMode.NONE

  # If you want use 'params' make sure that there is another mechanism to hide
  # these variables from the optimizer.
  quant_collection: str = 'aqt'

  assert_eqn: None | str = None
  assert_lhs_shape: None | utils.ShapeTemplate = None
  assert_rhs_shape: None | utils.ShapeTemplate = None
  tile_sizes: None | tiled_dot_general.EinsumTileSizes = None
  tiling_fn: None | EinsumTilingFn = None

  # If set to True, use the current Freezer. Otherwise, use the new
  # QTensorFreezer.
  use_legacy_freezer: bool = True

  @nn.compact
  def __call__(
      self,
      eqn,
      lhs_g: jnp.ndarray | aqt_tensor.QTensor,
      rhs_g: jnp.ndarray | aqt_tensor.QTensor,
  ):
    if self.assert_eqn is not None:
      utils.assert_eq(eqn, self.assert_eqn, 'einsum_eqn')
    if self.assert_lhs_shape is not None:
      utils.assert_shape(lhs_g.shape, self.assert_lhs_shape, 'lhs.shape')
    if self.assert_rhs_shape is not None:
      utils.assert_shape(rhs_g.shape, self.assert_rhs_shape, 'rhs.shape')

    cfg = self.cfg
    lhs_is_qt = isinstance(lhs_g, aqt_tensor.QTensor)
    rhs_is_qt = isinstance(rhs_g, aqt_tensor.QTensor)
    msg = 'Aqt config is None but inputs to AqtEinsum are QTensor.'
    assert not ((lhs_is_qt or rhs_is_qt) and cfg is None), msg
    # when inputs are qtensor, xhs_in is a dummy input that will be consumed by
    # lax einsum API, but it is not used for computation in aqt_dg because it
    # will be overwritten by get_tensor()
    # TODO(lew): We can pass QTensor to lax_numpy._einsum if we add some
    # specific methods to QTensor.
    lhs_in = jnp.zeros_like(lhs_g.qvalue) if lhs_is_qt else lhs_g
    rhs_in = jnp.zeros_like(rhs_g.qvalue) if rhs_is_qt else rhs_g

    # Set the types of dummy input to the same as original input, to prevent it
    # from being rejected by assertions in aqt_dot_general.py, line 522-526 and
    # 414.
    # TODO: b/322111904 - Handle this in more proper way.
    # We hand-hold int4 because promote_dtype(int4, x) fails.
    # (To avoid unintended promotion, 4-bit integers do not support
    # implicit promotion.)
    if lhs_in.dtype == jnp.int4:
      lhs_in = jnp.float32(lhs_in)
    if rhs_in.dtype == jnp.int4:
      rhs_in = jnp.float32(rhs_in)
    if lhs_in.dtype != jnp.int4 and rhs_in.dtype != jnp.int4:
      lhs_in, rhs_in = nn.dtypes.promote_dtype(lhs_in, rhs_in)

    # yes_swap = whether einsum swaps [lhs,rhs] when passing them to dot_general
    einsum = functools.partial(aqt_dot_general.einsum, eqn=eqn)
    a = jax.make_jaxpr(einsum)(lhs=lhs_in, rhs=rhs_in)
    [lhs_g_id, rhs_g_id] = a.eqns[0].invars
    [lhs_l_id, rhs_l_id] = a.jaxpr.invars
    not_swap = lhs_g_id == lhs_l_id and rhs_g_id == rhs_l_id
    yes_swap = lhs_g_id == rhs_l_id and rhs_g_id == lhs_l_id
    assert not_swap != yes_swap

    prng_name = self.prng_name

    lhs_quant_mode = self.lhs_quant_mode
    lhs_init = self.lhs_init
    lhs_axis_metadata_wrapper = self.lhs_axis_metadata_wrapper
    lhs_scale_init = self.lhs_scale_init
    lhs_var_name = self.lhs_var_name
    lhs_qtensor = lhs_g if lhs_is_qt else None

    rhs_quant_mode = self.rhs_quant_mode
    rhs_init = self.rhs_init
    rhs_axis_metadata_wrapper = self.rhs_axis_metadata_wrapper
    rhs_scale_init = self.rhs_scale_init
    rhs_var_name = self.rhs_var_name
    rhs_qtensor = rhs_g if rhs_is_qt else None

    lhs_freeze_mode = self.lhs_freeze_mode
    rhs_freeze_mode = self.rhs_freeze_mode

    quant_collection = self.quant_collection
    tiling_config = None
    if self.tile_sizes is not None:
      tiling_config = tiled_dot_general.Cfg.from_einsum(eqn, self.tile_sizes)
    elif self.tiling_fn is not None:
      tiling_config = self.tiling_fn(eqn, lhs_in, rhs_in)

    if yes_swap:
      if cfg is not None:
        cfg = copy.deepcopy(cfg)
        cfg.fwd.lhs, cfg.fwd.rhs = cfg.fwd.rhs, cfg.fwd.lhs
        cfg.fwd.dg_quantizer.swap_lhs_and_rhs()
        cfg.dlhs, cfg.drhs = cfg.drhs, cfg.dlhs
      lhs_quant_mode, rhs_quant_mode = rhs_quant_mode, lhs_quant_mode
      lhs_init, rhs_init = rhs_init, lhs_init
      lhs_scale_init, rhs_scale_init = rhs_scale_init, lhs_scale_init
      lhs_var_name, rhs_var_name = rhs_var_name, lhs_var_name
      lhs_is_qt, rhs_is_qt = rhs_is_qt, lhs_is_qt
      lhs_qtensor, rhs_qtensor = rhs_qtensor, lhs_qtensor
      lhs_axis_metadata_wrapper, rhs_axis_metadata_wrapper = (
          rhs_axis_metadata_wrapper,
          lhs_axis_metadata_wrapper,
      )
      lhs_freeze_mode, rhs_freeze_mode = rhs_freeze_mode, lhs_freeze_mode
      if tiling_config is not None:
        tiling_config = tiled_dot_general.Cfg(
            lhs=tiling_config.rhs, rhs=tiling_config.lhs
        )

    aqt_dg = AqtDotGeneral(
        cfg=cfg,
        prng_name=prng_name,
        lhs_quant_mode=lhs_quant_mode,
        # when passing pre-computed qtensor as inputs, apply_quant_mode flag
        # should be set to False so that Freezer will not be set to overwrite
        # the qtensor passed to dg.
        lhs_apply_quant_mode=not lhs_is_qt,  # Freezer not used if lhs is qt
        lhs_init=lhs_init,
        lhs_axis_metadata_wrapper=lhs_axis_metadata_wrapper,
        lhs_scale_init=lhs_scale_init,
        lhs_var_name=lhs_var_name,
        lhs_qtensor=lhs_qtensor,
        rhs_quant_mode=rhs_quant_mode,
        rhs_apply_quant_mode=not rhs_is_qt,  # Freezer not used if rhs is qt
        rhs_init=rhs_init,
        rhs_axis_metadata_wrapper=rhs_axis_metadata_wrapper,
        rhs_scale_init=rhs_scale_init,
        rhs_var_name=rhs_var_name,
        rhs_qtensor=rhs_qtensor,
        quant_collection=quant_collection,
        tiling_cfg=tiling_config,
        use_legacy_freezer=self.use_legacy_freezer,
        lhs_freeze_mode=lhs_freeze_mode,
        rhs_freeze_mode=rhs_freeze_mode,
    )
    return einsum(lhs=lhs_in, rhs=rhs_in, dg=aqt_dg)


class AqtConvGeneralDilated(nn.Module):
  """A layer that can be injected into flax.nn.Dense, etc."""

  cfg: None | aqt_dot_general.DotGeneral = None
  prng_name: None | str = 'params'

  # TODO(lew): split out separate class for each side.
  # Quant mode determines whether flax variables are created to store quantized
  # inputs. Refer to the Freezer doc str to see variable creation in each mode.
  lhs_quant_mode: QuantMode = QuantMode.TRAIN
  # apply_quant_mode determines if using Freezer in cfg.get/set_tensor
  lhs_apply_quant_mode: bool = True
  lhs_var_name: str = 'qlhs'
  lhs_qtensor: None | aqt_tensor.QTensor = None

  rhs_quant_mode: QuantMode = QuantMode.TRAIN
  rhs_apply_quant_mode: bool = True
  rhs_var_name: str = 'qrhs'
  rhs_qtensor: None | aqt_tensor.QTensor = None

  # Variables only for the new Freezer.
  lhs_axis_metadata_wrapper: None | AxisMetadataWrapper = None
  rhs_axis_metadata_wrapper: None | AxisMetadataWrapper = None

  # Freeze mode. Set as FreezerMode.CALIBRATION to store only scales; set as
  # CALIBRATION_AND_VALUE to store both scales and quantized values.
  lhs_freeze_mode: FreezerMode = FreezerMode.NONE
  rhs_freeze_mode: FreezerMode = FreezerMode.NONE

  # If you want use 'params' make sure that there is another mechanism to hide
  # these variables from the optimizer.
  quant_collection: str = 'aqt'

  def make_aqt_cg(
      self,
  ):
    if self.cfg is None:
      return jax.lax.conv_general_dilated

    cfg = copy.deepcopy(self.cfg)
    rhs_qm = self.rhs_quant_mode
    lhs_qm = self.lhs_quant_mode

    assert isinstance(
        cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
    )
    lhs_init_wrapper = functools.partial(
        _freezer_qtensor_init_wrapper,
        contracting_axis=[],
        axis_metadata_wrapper=self.lhs_axis_metadata_wrapper,
        tile_map=None,
    )
    rhs_init_wrapper = functools.partial(
        _freezer_qtensor_init_wrapper,
        contracting_axis=[],
        axis_metadata_wrapper=self.rhs_axis_metadata_wrapper,
        tile_map=None,
    )
    lhs_freezer = general_freezer.Freezer(
        name=self.lhs_var_name,
        mode=_QUANT_TO_FREEZER_MODE[lhs_qm],
        collection=self.quant_collection,
        axis_metadata_wrapper=lhs_init_wrapper,
    )
    rhs_freezer = general_freezer.Freezer(
        name=self.rhs_var_name,
        mode=_QUANT_TO_FREEZER_MODE[rhs_qm],
        collection=self.quant_collection,
        axis_metadata_wrapper=rhs_init_wrapper,
    )

    prng_name = self.prng_name
    key = self.make_rng(prng_name) if prng_name is not None else None

    cfg = config.set_context(
        cfg,
        key,
        train_step=None,
        lhs_quant_mode=self.lhs_quant_mode,
        rhs_quant_mode=self.rhs_quant_mode,
    )

    def ret_cg(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_dilation,
        rhs_dilation,
        dimension_numbers,
        feature_group_count,
        batch_group_count,
        precision,
        preferred_element_type,
    ):
      # TODO(yichizh): asserting xhs dtype only when apply_quant_mode=False
      # and cfg.get_qtensor() is None
      msg = 'AQT is not yet optimized to accept quantized types directly. '
      msg += f'lhs.dtype: {lhs.dtype}, rhs.dtype: {rhs.dtype}'
      assert lhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
      assert rhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg

      cfg.assert_config_validity()

      # Getter
      lhs_apply_quant_mode = self.lhs_apply_quant_mode
      rhs_apply_quant_mode = self.rhs_apply_quant_mode
      lhs_qt = lhs_freezer.get() if lhs_apply_quant_mode else self.lhs_qtensor
      rhs_qt = rhs_freezer.get() if rhs_apply_quant_mode else self.rhs_qtensor

      cfg.apply_custom_vjp_on_jax = False
      aqt_conv = aqt_conv_general.make_conv_general_dilated_with_qt(cfg.fwd)
      out, (out_lhs_qt, out_rhs_qt) = aqt_conv(
          lhs,
          rhs,
          window_strides,
          padding,
          lhs_qt,
          rhs_qt,
          lhs_dilation,
          rhs_dilation,
          dimension_numbers,
          feature_group_count,
          batch_group_count,
          precision,
          preferred_element_type
          )

      # Remove qvalue of the activation to not to store it in Freezer.
      match self.lhs_freeze_mode:
        case FreezerMode.NONE:
          if self.lhs_apply_quant_mode and self.lhs_quant_mode in {
              QuantMode.CONVERT,
              QuantMode.SERVE,
          }:
            raise ValueError('Freezer is used with Freezer mode NONE.')
        case FreezerMode.CALIBRATION:
          out_lhs_qt = out_lhs_qt.without_qvalue()
        case FreezerMode.CALIBRATION_AND_VALUE:
          pass
        case _:
          raise ValueError('Unknown freeze mode: %s' % self.lhs_freeze_mode)

      match self.rhs_freeze_mode:
        case FreezerMode.NONE:
          if self.rhs_apply_quant_mode and self.rhs_quant_mode in {
              QuantMode.CONVERT,
              QuantMode.SERVE,
          }:
            raise ValueError('Freezer is used with Freezer mode NONE.')
        case FreezerMode.CALIBRATION:
          out_rhs_qt = out_rhs_qt.without_qvalue()
        case FreezerMode.CALIBRATION_AND_VALUE:
          pass
        case _:
          raise ValueError('Unknown freeze mode: %s' % self.rhs_freeze_mode)

      if self.lhs_apply_quant_mode:
        lhs_freezer.set(out_lhs_qt)
      if self.rhs_apply_quant_mode:
        rhs_freezer.set(out_rhs_qt)

      return out

    return ret_cg

  @nn.compact
  def __call__(
      self,
      lhs,
      rhs,
      window_strides: Sequence[int],
      padding: str | Sequence[tuple[int, int]],
      lhs_dilation: None | Sequence[int] = None,
      rhs_dilation: None | Sequence[int] = None,
      dimension_numbers: (
          None
          | jax.lax.ConvGeneralDilatedDimensionNumbers
          | tuple[str, str, str]
      ) = None,
      feature_group_count: int = 1,
      batch_group_count: int = 1,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type=None,
  ):

    return self.make_aqt_cg()(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_dilation,
        rhs_dilation,
        dimension_numbers,
        feature_group_count,
        batch_group_count,
        precision,
        preferred_element_type,
    )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multiple sized model with varied quantization and sparsity.

Quantization is enabled since starting
Sparsity is one shot pruning, updated and applied at 100K steps.

gxm third_party/py/aqt/jax_legacy/jax/wmt_mlperf/google/xm_launch.py
--xm_resource_alloc=group:peace/babelfish-device
--hparams_config_dir=third_party/py/aqt/jax_legacy/jax/wmt_mlperf/hparams_configs/experimental/tmlr
--hparams_config_filename=sparsity_decay_nm_quant_wa.py --cell=yo
--sparsity_start_step=10000 --sparsity_update_freq=1000
--name="decay_sparsity_quant_sweep_2"
"""

import copy

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.tmlr import full_model_bfloat16
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.tmlr import small_model_bfloat16
import ml_collections


def get_config():
  """Sweep for multiple sized model with varied quantization and sparsity."""
  sweep_config = ml_collections.ConfigDict()
  base_configs = [
      full_model_bfloat16.get_config(
          quant_target=base_config.QuantTarget.WEIGHTS_AND_AUTO_ACTS
      ),
      small_model_bfloat16.get_config(
          quant_target=base_config.QuantTarget.WEIGHTS_AND_AUTO_ACTS
      ),
  ]

  configs = []
  for model_config in base_configs:
    for prec in [8, 4, 2]:
      for prune_rate in [(2, 4), (1, 4)]:
        config = copy.deepcopy(model_config)

        config.quant_type = 'aqt'

        # mlp + attention weights + acts  quantization
        config.dense.weight_prec = prec
        config.dense.quant_act.prec = prec

        config.num_train_steps = 100000

        # mlp weights sparsity
        config.mlp_block.dense_1.weight_sparsity.type = 'STRUCTURED_NM'
        config.mlp_block.dense_1.weight_sparsity.prune_rate = prune_rate
        config.mlp_block.dense_1.weight_sparsity.structure_decay = True

        config.mlp_block.dense_2.weight_sparsity.type = 'STRUCTURED_NM'
        config.mlp_block.dense_2.weight_sparsity.prune_rate = prune_rate
        config.mlp_block.dense_2.weight_sparsity.structure_decay = True

        # attn_weights sparsity
        config.attention.dense_kqv.weight_sparsity.type = 'STRUCTURED_NM'
        config.attention.dense_kqv.weight_sparsity.prune_rate = prune_rate
        config.attention.dense_kqv.weight_sparsity.structure_decay = True

        config.attention.dense_out.weight_sparsity.type = 'STRUCTURED_NM'
        config.attention.dense_out.weight_sparsity.prune_rate = prune_rate
        config.attention.dense_out.weight_sparsity.structure_decay = True

        config.metadata.hyper_str = f'{config.metadata.hyper_str}_decay_NM({prune_rate})_prec({prec}_100000steps)'
        configs.append(config)

  sweep_config.configs = configs
  return sweep_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mnist example."""

from absl import app
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2.examples import flax_e2e_model
from aqt.jax.v2.extensions.gptq import gptq_dot_general_quantizer
from aqt.jax.v2.numerics import int_numerics

create_train_state = flax_e2e_model.create_train_state
train_and_evaluate = flax_e2e_model.train_and_evaluate
train_epoch = flax_e2e_model.train_epoch
calibrate = flax_e2e_model.calibrate
calibration_conversion = flax_e2e_model.calibration_conversion
calibrate_epoch = flax_e2e_model.calibrate_epoch
serve = flax_e2e_model.serve
serving_conversion = flax_e2e_model.serving_conversion


def update_cfg_with_gptq(aqt_cfg: aqt_dot_general.DotGeneral) -> None:
  """Updates aqt_cfg for GPTQ.

  Args:
    aqt_cfg: aqt_dot_general configuration.
  """

  assert isinstance(
      aqt_cfg.fwd.dg_quantizer, aqt_dot_general.DefaultDotGeneralQuantizer
  )
  assert isinstance(
      aqt_cfg.fwd.dg_quantizer.lhs.numerics, int_numerics.IntSymmetric
  )
  assert isinstance(
      aqt_cfg.fwd.dg_quantizer.rhs.numerics, int_numerics.IntSymmetric
  )
  lhs_bits = aqt_cfg.fwd.dg_quantizer.lhs.numerics.bits
  rhs_bits = aqt_cfg.fwd.dg_quantizer.rhs.numerics.bits
  lhs = aqt_quantizer.quantizer_make(lhs_bits, initialize_calibration=False)
  rhs = aqt_quantizer.quantizer_make(rhs_bits, initialize_calibration=False)
  lhs_mid = aqt_quantizer.quantizer_make(lhs_bits, initialize_calibration=False)
  rhs_mid = aqt_quantizer.quantizer_make(rhs_bits, initialize_calibration=False)
  gptq_dg_quantizer = gptq_dot_general_quantizer.GptqDotGeneralQuantizer(
      lhs=lhs,
      rhs=rhs,
      lhs_mid=lhs_mid,
      rhs_mid=rhs_mid,
      sharding_axes=None,
      quant_collection='gptq',
  )

  aqt_cfg.fwd.dg_quantizer = gptq_dg_quantizer


def main(argv):
  del argv

  # 1. TRAIN.
  aqt_cfg_dg = aqt_config.fully_quantized(fwd_bits=8, bwd_bits=8)
  num_epochs = 1
  workdir = '/tmp/aqt_mnist_example'
  state = train_and_evaluate(
      num_epochs, workdir, aqt_cfg_dg=aqt_cfg_dg
  )

  # 2. Apply GPTQ Calibration. (Hinv collection).
  update_cfg_with_gptq(state.cnn_train.aqt_cfg)
  update_cfg_with_gptq(state.cnn_eval.aqt_cfg)
  state = calibrate(state, calibration_steps=10)

  # 3. CONVERT & SERVE.
  loss = serve(state, weight_only=True)
  print('serve loss on sample ds: {}'.format(loss))


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Minimally-sized unquantized model.

The minimal model has the smallest legal values for all model size parameters.

Useful for testing that training the model doesn't generate errors.
"""

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config


def get_config(quant_target=base_config.QuantTarget.NONE):
  """Returns configuration for a minimal transformer model."""
  config = base_config.get_config(quant_target=quant_target, n_layers=1)
  config.num_train_steps = 1
  model = config.model_hparams
  model.emb_dim = 1
  model.num_heads = 1
  model.qkv_dim = 1
  model.mlp_dim = 1
  return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for flax e2e model."""

import copy

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import config
from aqt.jax.v2 import utils
from aqt.jax.v2.extensions.gptq.examples import gptq_flax_e2e_model
import jax
import jax.numpy as jnp


def _dummy_dataset(ds_size, image_rng, label_rng):
  return {
      "image": jax.random.uniform(key=image_rng, shape=(ds_size, 28, 28, 1)),
      "label": jax.random.randint(
          key=label_rng, shape=(ds_size,), minval=0, maxval=10
      ),
  }


class GptqTest(parameterized.TestCase):

  def test_gptq(self):
    aqt_cfg_dg = config.config_v4()

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng, label_rng = jax.random.split(rng, 3)
    rng, input_rng = jax.random.split(rng)
    rng, calibration_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 64
    batch_size = 8
    ds = _dummy_dataset(ds_size, image_rng, label_rng)

    # Stage 1: regular training
    state = gptq_flax_e2e_model.create_train_state(
        init_rng, aqt_cfg_dg=aqt_cfg_dg)

    state, _, _ = gptq_flax_e2e_model.train_epoch(
        state, ds, batch_size, rng=input_rng
    )

    # Stage 2: Calibration.
    gptq_flax_e2e_model.update_cfg_with_gptq(state.cnn_train.aqt_cfg_dg)
    gptq_flax_e2e_model.update_cfg_with_gptq(state.cnn_eval.aqt_cfg_dg)

    calibrate_f, model_calibrate = gptq_flax_e2e_model.calibration_conversion(
        state
    )

    calibration_steps = 4
    calibrated_params = gptq_flax_e2e_model.calibrate_epoch(
        calibrate_f,
        model_calibrate,
        ds,
        batch_size,
        rng=calibration_rng,
        calibration_steps=calibration_steps,
    )
    calibration_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), calibrated_params
    )
    dtype = jnp.dtype
    expected_gptq_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "GptqHinvCollector_0": {
                    "collected_hinv": (dtype("float32"), (5, 5)),
                    "num_calibrated_batches": (dtype("int32"), ()),
                }
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "GptqHinvCollector_0": {
                    "collected_hinv": (dtype("float32"), (1568, 1568)),
                    "num_calibrated_batches": (dtype("int32"), ()),
                }
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "GptqHinvCollector_0": {
                    "collected_hinv": (dtype("float32"), (128, 128)),
                    "num_calibrated_batches": (dtype("int32"), ()),
                }
            }
        },
    }

    utils.test_pprint_eq(expected_gptq_pytree, calibration_pytree["gptq"])

    # The count number should be equal to the number of calibration.
    einsum_params = calibrated_params["gptq"]["AqtEinsum_0"]["AqtDotGeneral_0"]
    gptq_collected = einsum_params["GptqHinvCollector_0"]
    collected_count = gptq_collected["num_calibrated_batches"]
    self.assertEqual(calibration_steps, collected_count)

    # Stage 3. Convert the calibrated checkpoint.
    state = state.replace(model=copy.deepcopy(calibrated_params))
    _, model_serving = gptq_flax_e2e_model.serving_conversion(state)
    dtype = jnp.dtype
    expected_dtype = dtype("int8")
    expected_aqt_pytree = {
        "AqtEinsum_0": {
            "AqtDotGeneral_0": {
                "qlhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 5, 10)),
                        scale=[(dtype("float32"), (2, 1, 10))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                },
            }
        },
        "Dense_0": {
            "AqtDotGeneral_0": {
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 1568, 256)),
                        scale=[(dtype("float32"), (2, 1, 256))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                }
            }
        },
        "Dense_1": {
            "AqtDotGeneral_0": {
                "qrhs": {
                    "frozen": aqt_tensor.QTensor(
                        qvalue=(expected_dtype, (2, 128, 10)),
                        scale=[(dtype("float32"), (2, 1, 10))],
                        scale_t=None,
                        bias=[],
                        dequant_dtype=dtype("float32"),
                    )
                }
            }
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )

    utils.test_pprint_eq(expected_aqt_pytree, serving_pytree["aqt"])

    # Since the GPTQ changes the weights to better quantize it, the logits
    # before and after the conversion should be different.
    # As a conclusion, we do not put the logits comparison tests here.


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Util functions for numerics in AQT v2."""

from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import fp8_numerics
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics


def get_numerics(
    bits: None | int | fp8_numerics.FP8Dtype, preserve_max_val=False
):
  """Get numerics object from number of bits."""
  if bits is None:
    effective_numerics = no_numerics.NoNumerics()
  elif bits in fp8_numerics.fp8_map.keys():
    exponent_bits, mantissa_bits = int(bits[1]), int(bits[3])
    effective_numerics = fp8_numerics.Fp8Numerics(
        exponent_bits=exponent_bits,
        mantissa_bits=mantissa_bits,
        dtype=fp8_numerics.fp8_map[bits],
    )
  else:
    pz = False if bits == 1 else True
    dtype = utils.infer_dtype_from_bits(bits) if pz else None
    effective_numerics = int_numerics.IntSymmetric(
        bits=bits,
        preserve_zero=pz,
        preserve_max_val=preserve_max_val,
        clip=True,
        round=True,
        noise_fn=None,
        clip_gradient=False,  # Can be disabled when using abs-max scaling.
        dtype=dtype,
    )
  return effective_numerics



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General util functions commonly used across different models."""


def get_fp_spec(sig_bit: int, exp_bit: int):
  """Create fp spec which defines precision for floating-point quantization.

  Args:
    sig_bit: the number of bits assigned for significand.
    exp_bit: the number of bits assigned for exponent.

  Returns:
    fp spec
  """
  exp_bound = 2**(exp_bit - 1) - 1
  prec = {'exp_min': -exp_bound, 'exp_max': exp_bound, 'sig_bits': sig_bit}
  return prec
